"""liquidctl driver for Corsair Platinum and PRO XT coolers.

Supported devices
-----------------

 - Corsair H110i GT

Supported features
------------------

 - general monitoring
 - pump speed control
 - fan speed control

Copyright Roberto Marques, Serphentas, and other contributors
SPDX-License-Identifier: GPL-3.0-or-later
"""

import logging
import re

from enum import Enum, unique

from liquidctl.error import NotSupportedByDevice, NotSupportedByDriver
from liquidctl.driver.usb import UsbHidDriver
from liquidctl.keyval import RuntimeStorage
from liquidctl.util import clamp, fraction_of_byte, u16le_from, normalize_profile

LOGGER = logging.getLogger(__name__)

_REPORT_LENGTH = 64
_PUMP_INDEX = 0x02

_COMMAND_FIRMWARE_ID = 0x01
_COMMAND_TEMP_READ = 0x0E
_COMMAND_FAN_SELECT = 0x10
_COMMAND_FAN_MODE = 0x12
_COMMAND_FAN_FIXED_PWM = 0x13
_COMMAND_FAN_FIXED_RPM = 0x14
_COMMAND_FAN_READ_RPM = 0x16
_COMMAND_FAN_MAX_RPM = 0x17
_COMMAND_FAN_RPM_TABLE = 0x19
_COMMAND_FAN_TEMP_TABLE = 0x1A

_OP_CODE_WRITE_ONE_BYTE = 0x06
_OP_CODE_READ_ONE_BYTE = 0x07
_OP_CODE_WRITE_TWO_BYTES = 0x08
_OP_CODE_READ_TWO_BYTES = 0x09
_OP_CODE_WRITE_THREE_BYTES = 0x0A
_OP_CODE_READ_THREE_BYTES = 0x0B

_PROFILE_LENGTH = 5
_CRITICAL_TEMPERATURE = 60

_PUMP_DEFAULT_QUIET = [0x2E, 0x09]
_PUMP_DEFAULT_EXTREME = [0x86, 0x0B]

_SEQUENCE_MIN = 1
_SEQUENCE_MAX = 255


@unique
class _FanMode(Enum):
    FIXED_DUTY = 0x02
    FIXED_RPM = 0x04
    CUSTOM_PROFILE = 0x0E

    @classmethod
    def _missing_(cls, value):
        LOGGER.debug("falling back to FIXED_DUTY for _FanMode(%s)", value)
        return _FanMode.FIXED_DUTY


@unique
class _PumpMode(Enum):
    QUIET = 0x08
    EXTREME = 0x0C

    @classmethod
    def _missing_(cls, value):
        LOGGER.debug("falling back to QUIET for _PumpMode(%s)", value)
        return _PumpMode.QUIET


def _sequence():
    """Return a generator that produces valid protocol sequence numbers.

    Sequence numbers start from 2 to 31, then rolling over to 1 and up again.
    """
    num = 1
    while True:
        yield (num % 31) + 1
        num += 1


def _prepare_profile(original):
    clamped = ((temp, clamp(duty, 0, 100)) for temp, duty in original)
    normal = normalize_profile(clamped, _CRITICAL_TEMPERATURE)
    missing = _PROFILE_LENGTH - len(normal)
    if missing < 0:
        raise ValueError(f"Too many points in profile (remove {-missing})")
    if missing > 0:
        normal += missing * [(_CRITICAL_TEMPERATURE, 100)]
    return normal


def _quoted(*names):
    return ", ".join(map(repr, names))


class Coolit(UsbHidDriver):
    """liquidctl driver for Corsair H110i GT cooler"""

    _MATCHES = [
        (
            0x1B1C,
            0x0C04,
            "Corsair H110i GT",
            {"has_pump": True, "fan_count": 2, "rgb_fans": False},
        ),
    ]

    def __init__(self, device, description, fan_count, rgb_fans, **kwargs):
        super().__init__(device, description, **kwargs)
        self._component_count = 1 + fan_count * rgb_fans
        self._fan_names = [f"fan{i + 1}" for i in range(fan_count)]

        # the following fields are only initialized in connect()
        self._data = None
        self._sequence = None

    def connect(self, **kwargs):
        """Connect to the device."""
        ret = super().connect(**kwargs)
        ids = f"vid{self.vendor_id:04x}_pid{self.product_id:04x}"
        loc = "loc" + "_".join(re.findall(r"\d+", self.address))
        self._data = RuntimeStorage(key_prefixes=[ids, loc])
        self._sequence = _sequence()
        return ret

    def initialize(self, pump_mode="quiet", **kwargs):
        """Initialize the device and set the pump mode

        The device should be initialized every time it is powered on, including when
        the system resumes from suspending to memory.

        Valid values for `pump_mode` are "quiet" and "extreme".
        Unconfigured fan channels may default to 100% duty.

        Returns a list of `(property, value, unit)` tuples.
        """
        if pump_mode not in ["quiet", "extreme"]:
            LOGGER.warning('pump mode must be either "quiet" or "extreme", falling back to "quiet"')
            pump_mode = "quiet"

        self._data.store("pump_mode", _PumpMode[pump_mode.upper()].value)

        res = self._send_command(
            self._build_data_package(_COMMAND_FIRMWARE_ID, _OP_CODE_READ_TWO_BYTES)
        )

        fw_version = (res[3] >> 4, res[3] & 0xF, res[2])
        return [("Firmware version", "%d.%d.%d" % fw_version, "")]

    def get_status(self, **kwargs):
        """Get a status report.

        Returns a list of `(property, value, unit)` tuples.
        """
        dataPackages = list()

        # temperature
        dataPackages.append(self._build_data_package(_COMMAND_TEMP_READ, _OP_CODE_READ_TWO_BYTES))

        # fan 1
        dataPackages.append(
            self._build_data_package(
                _COMMAND_FAN_SELECT, _OP_CODE_WRITE_ONE_BYTE, params=bytes([0])
            )
        )
        dataPackages.append(
            self._build_data_package(_COMMAND_FAN_READ_RPM, _OP_CODE_READ_TWO_BYTES)
        )

        # fan 2
        dataPackages.append(
            self._build_data_package(
                _COMMAND_FAN_SELECT, _OP_CODE_WRITE_ONE_BYTE, params=bytes([1])
            )
        )
        dataPackages.append(
            self._build_data_package(_COMMAND_FAN_READ_RPM, _OP_CODE_READ_TWO_BYTES)
        )

        # pump
        dataPackages.append(
            self._build_data_package(
                _COMMAND_FAN_SELECT, _OP_CODE_WRITE_ONE_BYTE, params=bytes([2])
            )
        )
        dataPackages.append(
            self._build_data_package(_COMMAND_FAN_READ_RPM, _OP_CODE_READ_TWO_BYTES)
        )

        res = self._send_commands(dataPackages)

        temp = res[3] + res[2] / 255

        return [
            ("Liquid temperature", temp, "°C"),
            ("Fan 1 speed", u16le_from(res, offset=8), "rpm"),
            ("Fan 2 speed", u16le_from(res, offset=14), "rpm"),
            ("Pump speed", u16le_from(res, offset=20), "rpm"),
        ]

    def set_fixed_speed(self, channel, duty, **kwargs):
        """Set fan or fans to a fixed speed duty.

        Valid channel values are "fanN", where N >= 1 is the fan number, and
        "fan", to simultaneously configure all fans.  Unconfigured fan channels
        may default to 100% duty.
        """
        for hw_channel in self._get_hw_fan_channels(channel):
            self._data.store(f"{hw_channel}_mode", _FanMode.FIXED_DUTY.value)
            self._data.store(f"{hw_channel}_duty", duty)
            LOGGER.info(f"setting {hw_channel} to duty mode")
        self._send_set_cooling()

    def set_speed_profile(self, channel, profile, **kwargs):
        """Set fan or fans to follow a speed duty profile.

        Valid channel values are "fanN", where N >= 1 is the fan number, and
        "fan", to simultaneously configure all fans.  Unconfigured fan channels
        may default to 100% duty.

        Up to seven (temperature, duty) pairs can be supplied in `profile`,
        with temperatures in Celsius and duty values in percentage.  The last
        point should set the fan to 100% duty cycle, or be omitted; in the
        latter case the fan will be set to max out at 60°C.
        """
        profile = list(profile)
        for hw_channel in self._get_hw_fan_channels(channel):
            self._data.store(f"{hw_channel}_mode", _FanMode.CUSTOM_PROFILE.value)
            self._data.store(f"{hw_channel}_profile", profile)
            LOGGER.info(f"setting {hw_channel} to profile mode")
        self._send_set_cooling()

    def set_color(self, channel, mode, colors, **kwargs):
        """Not supported by this driver."""
        raise NotSupportedByDriver()

    def set_screen(self, channel, mode, value, **kwargs):
        """Not supported by this device."""
        raise NotSupportedByDevice()

    def _get_hw_fan_channels(self, channel):
        channel = channel.lower()
        if channel == "fan":
            return self._fan_names
        if channel in self._fan_names:
            return [channel]
        raise ValueError(f"Unknown channel, should be one of: {_quoted('fan', *self._fan_names)}")

    def _build_data_package(self, command, opCode, params=None):
        if params:
            buf = bytearray(3 + len(params))
            buf[3 : 3 + len(params)] = params
        else:
            buf = bytearray(3)

        buf[0] = next(self._sequence)
        buf[1] = opCode
        buf[2] = command

        return buf

    def _send_commands(self, dataPackages):
        buf = bytearray(_REPORT_LENGTH)

        startIndex = 1
        for dataPackage in dataPackages:
            buf[startIndex : startIndex + len(dataPackage)] = dataPackage
            startIndex += len(dataPackage)

        buf[0] = startIndex - 1

        return self._send_buffer(buf)

    def _send_command(self, dataPackage):
        buf = bytearray(_REPORT_LENGTH)

        buf[0] = len(dataPackage)
        buf[1:] = dataPackage

        return self._send_buffer(buf)

    def _send_buffer(self, buf):
        self.device.clear_enqueued_reports()
        self.device.write(buf)
        buf = bytes(self.device.read(_REPORT_LENGTH))
        return buf

    def _send_set_cooling(self):
        for fan in self._fan_names:
            fanIndex = 0 if fan == "fan1" else 1

            mode = _FanMode(self._data.load(f"{fan}_mode", of_type=int))

            if mode is _FanMode.FIXED_DUTY:
                stored = self._data.load(f"{fan}_duty", of_type=int, default=100)
                duty = clamp(stored, 0, 100)
                self._send_command(
                    self._build_data_package(
                        _COMMAND_FAN_SELECT, _OP_CODE_WRITE_ONE_BYTE, params=bytes([fanIndex])
                    )
                )
                self._send_command(
                    self._build_data_package(
                        _COMMAND_FAN_MODE, _OP_CODE_WRITE_ONE_BYTE, params=bytes([mode.value])
                    )
                )
                self._send_command(
                    self._build_data_package(
                        _COMMAND_FAN_FIXED_PWM,
                        _OP_CODE_WRITE_ONE_BYTE,
                        params=bytes([fraction_of_byte(percentage=duty)]),
                    )
                )
                LOGGER.info("setting %s to %d%% duty cycle", fan, duty)

            elif mode is _FanMode.CUSTOM_PROFILE:
                stored = self._data.load(f"{fan}_profile", of_type=list, default=[])
                profile = _prepare_profile(stored)  # ensures correct len(profile)
                pairs = ((temp, fraction_of_byte(percentage=duty)) for temp, duty in profile)

                # "magical" 0x0A in front of curve definition packages
                fanTemperatureData = [0x0A]
                fanDutyData = [0x0A]

                # get max RPM for current fan
                dataPackages = []
                dataPackages.append(
                    self._build_data_package(
                        _COMMAND_FAN_SELECT, _OP_CODE_WRITE_ONE_BYTE, params=bytes([0])
                    )
                )
                dataPackages.append(
                    self._build_data_package(_COMMAND_FAN_MAX_RPM, _OP_CODE_READ_TWO_BYTES)
                )
                max_rpm = u16le_from(self._send_commands(dataPackages), offset=4)

                for temp, duty in profile:
                    fanTemperatureData.append(0x00)
                    fanTemperatureData.append(temp)

                    rpm = duty * max_rpm / 100
                    fanDutyData.append(int(rpm % 255))
                    fanDutyData.append(int(rpm - (rpm % 255)) >> 8)

                # select fan to customize
                self._send_command(
                    self._build_data_package(
                        _COMMAND_FAN_SELECT, _OP_CODE_WRITE_ONE_BYTE, params=bytes([fanIndex])
                    )
                )

                # Change mode to custom Profile
                self._send_command(
                    self._build_data_package(
                        _COMMAND_FAN_MODE, _OP_CODE_WRITE_ONE_BYTE, params=bytes([mode.value])
                    )
                )

                # Send duty cycle Profile
                self._send_command(
                    self._build_data_package(
                        _COMMAND_FAN_RPM_TABLE,
                        _OP_CODE_WRITE_THREE_BYTES,
                        params=bytes(fanDutyData),
                    )
                )

                # Send temperature profile
                self._send_command(
                    self._build_data_package(
                        _COMMAND_FAN_TEMP_TABLE,
                        _OP_CODE_WRITE_THREE_BYTES,
                        params=bytes(fanTemperatureData),
                    )
                )

                LOGGER.info("setting %s to follow profile %r", fan, profile)
            else:
                raise ValueError(f"Unsupported fan {mode}")

        pump_mode = _PumpMode(self._data.load("pump_mode", of_type=int))

        self._send_commands(
            [
                self._build_data_package(
                    _COMMAND_FAN_SELECT, _OP_CODE_WRITE_ONE_BYTE, params=bytes([_PUMP_INDEX])
                ),
                self._build_data_package(
                    _COMMAND_FAN_FIXED_RPM,
                    _OP_CODE_WRITE_TWO_BYTES,
                    params=bytes(
                        _PUMP_DEFAULT_QUIET
                        if pump_mode == _PumpMode.QUIET
                        else _PUMP_DEFAULT_EXTREME
                    ),
                ),
            ]
        )

        LOGGER.info("setting pump mode to %s", pump_mode.name.lower())



"""liquidctl driver for NZXT Control Hub.

Copyright Joseph Livecchi, and contributors
SPDX-License-Identifier: GPL-3.0-or-later
"""

from liquidctl.driver.smart_device import _BaseSmartDevice
from liquidctl.util import Hue2Accessory, HUE2_MAX_ACCESSORIES_IN_CHANNEL

_MIN_DUTY = 0
_MAX_DUTY = 100


class ControlHub(_BaseSmartDevice):
    """NZXT Control Hub RGB and fan controller."""

    _MATCHES = [
        (0x1E71, 0x2022, "NZXT Control Hub", {"speed_channel_count": 5, "color_channel_count": 5}),
    ]

    _MAX_READ_ATTEMPTS = 12
    _READ_LENGTH = 64
    _WRITE_LENGTH = 64

    # Color modes with extended packet configuration
    # Format: (mode_byte, variant_byte, moving_byte, min_colors, max_colors, packet_config)
    # packet_config only needs to specify overrides; defaults are applied in _build_mode_packet
    _COLOR_MODES = {
        "off": (0x00, 0x00, 0x00, 0, 0, {"type": "fixed"}),
        "fixed": (0x00, 0x00, 0x00, 1, 1, {"type": "fixed"}),
        "fading": (
            0x01,
            0x00,
            0x00,
            1,
            8,
            {
                "speed_map": "fading",
                "has_colors": True,
                "color_count_pos": 56,
                "footer": [0x08, 0x18, 0x03, 0x00, 0x00],
            },
        ),
        "spectrum-wave": (
            0x02,
            0x00,
            0x00,
            0,
            0,
            {"has_direction": True, "footer": [0x00, 0x00, 0x12, 0x03, 0x00, 0x00]},
        ),
        "covering-marquee": (
            0x04,
            0x00,
            0x00,
            1,
            8,
            {
                "has_direction": True,
                "has_colors": True,
                "pad_to": 55,
                "direction_pos": 55,
                "color_count_pos": 56,
                "footer": [0x00, 0x18, 0x03, 0x00, 0x00],
            },
        ),
        "super-rainbow": (
            0x0C,
            0x00,
            0x00,
            0,
            0,
            {
                "has_direction": True,
                "header_padding": [0x00, 0x00],
                "footer": [0x00, 0x18, 0x03, 0x00, 0x00],
            },
        ),
    }

    # Speed mappings for different mode types
    _SPEED_MAPS = {
        "standard": {  # Used by spectrum-wave, marquee, rainbow modes
            "slowest": [0x5E, 0x01],
            "slower": [0x2C, 0x01],
            "normal": [0xFA, 0x00],
            "faster": [0x96, 0x00],
            "fastest": [0x50, 0x00],
        },
        "fading": {  # Used by fading mode
            "slowest": [0x50, 0x00],
            "slower": [0x3C, 0x00],
            "normal": [0x28, 0x00],
            "faster": [0x14, 0x00],
            "fastest": [0x0A, 0x00],
        },
    }

    # Channel ID to byte mapping
    _CHANNEL_BYTE_MAP = {0: 0x02, 1: 0x04, 2: 0x06, 3: 0x08, 4: 0x10}

    def __init__(self, device, description, speed_channel_count, color_channel_count, **kwargs):
        """Instantiate a driver with a device handle."""
        speed_channels = {
            f"fan{i + 1}": (i, _MIN_DUTY, _MAX_DUTY) for i in range(speed_channel_count)
        }
        color_channels = {f"led{i + 1}": i for i in range(color_channel_count)}
        if color_channels:
            color_channels["sync"] = 0xFF  # Special value for all channels
        super().__init__(device, description, speed_channels, color_channels)

    def initialize(self, **kwargs):
        """Initialize the device and the driver.

        Connected fans and LED accessories are detected.

        This method should be called every time the systems boots, resumes from
        a suspended state, or if the device has just been (re)connected.  In
        those scenarios, no other method, except `connect()` or `disconnect()`,
        should be called until the device and driver has been (re-)initialized.

        Returns None or a list of `(property, value, unit)` tuples, similarly
        to `get_status()`.
        """

        self.device.clear_enqueued_reports()

        update_interval = (lambda secs: 1 + round((secs - 0.5) / 0.25))(0.5)  # see issue #128
        self._write([0x60, 0x02, 0x01, 0xE8, update_interval, 0x01, 0xE8, update_interval])
        self._write([0x66, 0x01])

        # request static infos
        self._write([0x10, 0x02])  # firmware info
        self._write([0x20, 0x03])  # lighting info
        ret = []

        def parse_firm_info(msg):
            fw = f"{msg[0x11]}.{msg[0x12]}.{msg[0x13]}"
            ret.append(("Firmware version", fw, ""))

        def parse_led_info(msg):
            channel_count = msg[14]
            offset = 15  # offset of first channel/first accessory
            for c in range(channel_count):
                for a in range(HUE2_MAX_ACCESSORIES_IN_CHANNEL):
                    accessory_id = msg[offset + c * HUE2_MAX_ACCESSORIES_IN_CHANNEL + a]
                    if accessory_id == 0:
                        break
                    ret.append((f"LED {c + 1} accessory {a + 1}", Hue2Accessory(accessory_id), ""))

        parsers = {b"\x11\x02": parse_firm_info}
        if self._color_channels:
            parsers[b"\x21\x03"] = parse_led_info
        self._read_until(parsers)
        return sorted(ret)

    def get_status(self, **kwargs):
        ret = []

        def parse_fan_info(msg):
            mode_offset = 16
            rpm_offset = 24
            duty_offset = 40
            noise_offset = 56
            raw_modes = [None, "DC", "PWM"]

            for i, _ in enumerate(self._speed_channels):
                mode = raw_modes[msg[mode_offset + i]]
                ret.append(
                    (f"Fan {i + 1} speed", msg[rpm_offset + 1] << 8 | msg[rpm_offset], "rpm")
                )
                ret.append((f"Fan {i + 1} duty", msg[duty_offset + i], "%"))
                ret.append((f"Fan {i + 1} control mode", mode, ""))
                rpm_offset += 2
            ret.append(("Noise level", msg[noise_offset], "dB"))

        self.device.clear_enqueued_reports()
        self._read_until({b"\x67\x02": parse_fan_info})
        return sorted(ret)

    def set_color(self, channel, mode, colors, speed="normal", direction="forward", **kwargs):
        """Set the color mode for a specific channel.

        Supported modes:
        - off: Turn off the channel
        - fixed: Set a fixed color
        - fading: Fade between colors
        - spectrum-wave: Spectrum wave effect
        - covering-marquee: Covering marquee effect
        - super-rainbow: Super rainbow effect
        """
        if channel not in self._color_channels:
            raise ValueError(f"invalid channel: {channel}")

        if mode not in self._COLOR_MODES:
            raise ValueError(
                f"invalid mode: {mode}, supported modes: {list(self._COLOR_MODES.keys())}"
            )

        # Unpack mode configuration (extended format with config dict)
        mode_byte, variant_byte, moving_byte, min_colors, max_colors, packet_config = (
            self._COLOR_MODES[mode]
        )

        # Convert colors to list if it's a map object
        colors_list = (
            list(colors)
            if hasattr(colors, "__iter__") and not isinstance(colors, (list, tuple))
            else colors
        )

        # Validate color count
        if len(colors_list) < min_colors:
            raise ValueError(f"mode {mode!r} requires at least {min_colors} colors")
        if len(colors_list) > max_colors:
            raise ValueError(f"mode {mode!r} supports at most {max_colors} colors")

        # Handle off mode
        if mode == "off":
            colors_list = [(0, 0, 0)]

        # Determine channel ID
        cid = self._color_channels[channel]

        # Set color mode and data
        if cid == 0xFF:  # sync mode - set all channels
            for ch_id in range(len(self._color_channels) - 1):  # -1 to exclude 'sync'
                self._set_channel_color_mode(
                    ch_id,
                    mode_byte,
                    variant_byte,
                    moving_byte,
                    colors_list,
                    speed,
                    direction,
                    packet_config,
                )
        else:
            self._set_channel_color_mode(
                cid,
                mode_byte,
                variant_byte,
                moving_byte,
                colors_list,
                speed,
                direction,
                packet_config,
            )

    def _set_channel_color_mode(
        self,
        channel_id,
        mode_byte,
        _variant_byte,
        _moving_byte,
        colors,
        speed,
        direction,
        packet_config,
    ):
        """Set color mode for a specific channel."""
        channel_byte = self._CHANNEL_BYTE_MAP.get(channel_id, 0x08)

        # Use packet_config to build the data
        data = self._build_mode_packet(
            channel_byte, mode_byte, colors, speed, direction, packet_config
        )

        # Pad and write
        padding = [0x00] * (self._WRITE_LENGTH - len(data))
        data.extend(padding)
        self._write(data)

        # Apply settings if required
        if packet_config.get("type") == "fixed":
            self._apply_color_settings(channel_byte)

    def _build_mode_packet(self, channel_byte, mode_byte, colors, speed, direction, config):
        """Build packet based on mode configuration from _COLOR_MODES.

        Applies sensible defaults for common fields:
        - type: 'animated' (unless specified as 'fixed')
        - base_cmd: [0x2a, 0x04] for animated, [0x26, 0x04] for fixed
        - speed_map: 'standard'
        - pad_to: 56
        - direction_pos: 56 (when has_direction is True)
        - led_count: 24 (for fixed type)
        """
        # Apply defaults
        packet_type = config.get("type", "animated")

        if packet_type == "fixed":
            # Fixed color mode defaults
            base_cmd = config.get("base_cmd", [0x26, 0x04])
            led_count = config.get("led_count", 24)

            data = base_cmd + [channel_byte, 0x00]
            if colors:
                r, g, b = colors[0]
                for _ in range(led_count):
                    data.extend([g, r, b])  # GRB order
            else:
                data.extend([0x00] * (led_count * 3))

        else:  # animated
            # Animated mode defaults
            base_cmd = config.get("base_cmd", [0x2A, 0x04])
            speed_map_name = config.get("speed_map", "standard")
            pad_to = config.get("pad_to", 56)
            _direction_pos = config.get("direction_pos", 56)

            data = base_cmd + [channel_byte, channel_byte, mode_byte]

            # Add speed bytes if speed_map is specified
            if speed_map_name:
                speed_map = self._SPEED_MAPS[speed_map_name]
                speed_bytes = speed_map.get(speed, speed_map["normal"])
                data.extend(speed_bytes)

            # Add header padding if specified
            if "header_padding" in config:
                data.extend(config["header_padding"])

            # Add colors if mode supports them
            if config.get("has_colors") and colors:
                for r, g, b in colors:
                    data.extend([g, r, b])  # GRB order

            # Pad to specified position
            while len(data) < pad_to:
                data.append(0x00)

            # Add direction byte if needed
            if config.get("has_direction"):
                direction_byte = 0x02 if direction == "backward" else 0x00
                data.append(direction_byte)

            # Add color count if needed
            if config.get("has_colors") and "color_count_pos" in config and colors:
                data.append(len(colors))

            # Add footer bytes if specified
            if "footer" in config:
                data.extend(config["footer"])

            # Pad to minimum length if no colors provided
            if not colors and config.get("has_colors"):
                while len(data) < 61:
                    data.append(0x00)

        return data

    def _apply_color_settings(self, channel_byte):
        """Apply/commit the color settings to the device."""
        # Based on USB capture analysis:
        # Format: 26 06 [channel] 00 01 00 00 18 00 00 80 00 32 00 00 01 00 00...
        data = [
            0x26,
            0x06,
            channel_byte,
            0x00,
            0x01,
            0x00,
            0x00,
            0x18,
            0x00,
            0x00,
            0x80,
            0x00,
            0x32,
            0x00,
            0x00,
            0x01,
            0x00,
            0x00,
        ]

        # Pad to full packet size
        padding = [0x00] * (self._WRITE_LENGTH - len(data))
        data.extend(padding)

        self._write(data)

    def _write_fixed_duty(self, cid, duty):
        # Fan channel passed as bitflag in last 3 bits of 3rd byte
        msg = [0x62, 0x01, 0x01 << cid, 0x00, 0x00, 0x00, 0x00, 0x00]
        # Duty percent in 4th, 5th, 6th, 7th and 8th bytes for, respectively, fan1, fan2, fan3, fan4
        # and fan5
        msg[cid + 3] = duty
        self._write(msg)

    def _read_until(self, parsers):
        for _ in range(self._MAX_READ_ATTEMPTS):
            msg = self.device.read(self._READ_LENGTH)
            prefix = bytes(msg[0:2])
            func = parsers.pop(prefix, None)
            if func:
                func(msg)
            if not parsers:
                return
        assert (
            False
        ), f"missing messages (attempts={self._MAX_READ_ATTEMPTS}, missing={len(parsers)})"



"""liquidctl driver for the ASUS Ryuo liquid coolers.

Copyright Bloodhundur and contributors
SPDX-License-Identifier: GPL-3.0-or-later
"""

import logging
from typing import List

from liquidctl.driver.usb import UsbHidDriver
from liquidctl.error import ExpectationNotMet
from liquidctl.util import clamp, rpadlist

_LOGGER = logging.getLogger(__name__)

_REPORT_LENGTH = 65
_PREFIX = 0xEC
_CMD_SET_FAN_SPEED = 0x2A

_REQUEST_GET_FIRMWARE = (0x82, 0x02)
_STATUS_FIRMWARE = "Firmware version"


class AsusRyuo(UsbHidDriver):
    """Driver for ASUS Ryuo I 240 (fan control only)."""

    _MATCHES = [
        (0x0B05, 0x1887, "ASUS Ryuo I 240", {}),
    ]

    def initialize(self, **kwargs):
        msg = self._request(*_REQUEST_GET_FIRMWARE)
        fw_string = bytes(msg[2:]).split(b"\x00")[0].decode("ascii", errors="ignore")
        return [
            (_STATUS_FIRMWARE, fw_string, ""),
        ]

    def get_status(self, **kwargs):
        _LOGGER.info("No status available")

    def set_fixed_speed(self, channel, duty, **kwargs):
        if channel not in ("fans", "fan"):
            raise ValueError("Only 'fans' channel is supported")
        duty = clamp(duty, 0, 100)
        _LOGGER.info(f"Setting fan speed to {duty}%")
        self._write([_PREFIX, _CMD_SET_FAN_SPEED, duty])

    def _request(self, request_header: int, response_header: int) -> List[int]:
        self.device.clear_enqueued_reports()
        self._write([_PREFIX, request_header])
        return self._read(response_header)

    def _read(self, expected_header=None) -> List[int]:
        msg = self.device.read(_REPORT_LENGTH)
        if msg[0] != _PREFIX:
            raise ExpectationNotMet("Unexpected report prefix")
        if expected_header is not None and msg[1] != expected_header:
            raise ExpectationNotMet("Unexpected report header")
        return msg

    def _write(self, data: List[int]):
        self.device.write(rpadlist(data, _REPORT_LENGTH, 0))



"""liquidctl driver for the ASUS Ryujin II and Ryujin III EXTREME / 360 liquid coolers.

Copyright Florian Freudiger and contributors
SPDX-License-Identifier: GPL-3.0-or-later
"""

import logging
from typing import List

from liquidctl.driver.usb import UsbHidDriver
from liquidctl.error import ExpectationNotMet, NotSupportedByDriver
from liquidctl.util import clamp, u16le_from, rpadlist, fraction_of_byte

_LOGGER = logging.getLogger(__name__)

_REPORT_LENGTH = 65
_PREFIX = 0xEC

# Requests and their response headers
_REQUEST_GET_FIRMWARE = (0x82, 0x02)
_REQUEST_GET_COOLER_STATUS = (0x99, 0x19)
_REQUEST_GET_COOLER_DUTY = (0x9A, 0x1A)
_REQUEST_GET_CONTROLLER_DUTY = (0xA1, 0x21)
_REQUEST_GET_CONTROLLER_SPEED = (0xA0, 0x20)

# Command headers that don't need a response
_CMD_SET_COOLER_SPEED = 0x1A
_CMD_SET_CONTROLLER_SPEED = 0x21

_STATUS_FIRMWARE = "Firmware version"
_STATUS_TEMPERATURE = "Liquid temperature"
_STATUS_PUMP_SPEED = "Pump speed"
_STATUS_PUMP_DUTY = "Pump duty"
_STATUS_COOLER_FAN_SPEED = "Pump fan speed"
_STATUS_COOLER_FAN_DUTY = "Pump fan duty"
_STATUS_CONTROLLER_FAN_SPEED = "External fan {} speed"
_STATUS_CONTROLLER_FAN_DUTY = "External fan duty"


class AsusRyujin(UsbHidDriver):
    """ASUS Ryujin II & Ryujin III Extreme / 360 / White Edition liquid coolers."""

    _MATCHES = [
        (
            0x0B05,
            0x1988,
            "ASUS Ryujin II 360",
            {
                "fan_count": 4,
                "pump_speed_offset": 5,
                "pump_fan_speed_offset": 7,
                "temp_offset": 3,
                "duty_channel": 0,
            },
        ),
        (
            0x0B05,
            0x1BCB,
            "ASUS Ryujin III Extreme",
            {
                "fan_count": 0,
                "pump_speed_offset": 7,
                "pump_fan_speed_offset": 10,
                "temp_offset": 5,
                "duty_channel": 1,
            },
        ),
        (
            0x0B05,
            0x1AA2,
            "ASUS Ryujin III 360",
            {
                "fan_count": 0,
                "pump_speed_offset": 7,
                "pump_fan_speed_offset": 10,
                "temp_offset": 5,
                "duty_channel": 1,
            },
        ),
        (
            0x0B05,
            0x1ADE,
            "ASUS Ryujin III EVA",
            {
                "fan_count": 0,
                "pump_speed_offset": 7,
                "pump_fan_speed_offset": 10,
                "temp_offset": 5,
                "duty_channel": 1,
            },
        ),
        (
            0x0B05,
            0x1ADA,
            "ASUS Ryujin III White",
            {
                "fan_count": 0,
                "pump_speed_offset": 7,
                "pump_fan_speed_offset": 10,
                "temp_offset": 5,
                "duty_channel": 1,
            },
        ),
    ]

    def __init__(
        self,
        device,
        description,
        fan_count,
        pump_speed_offset,
        pump_fan_speed_offset,
        temp_offset,
        duty_channel,
        **kwargs,
    ):
        super().__init__(device, description, **kwargs)

        self._fan_count = fan_count
        self._pump_speed_offset = pump_speed_offset
        self._pump_fan_speed_offset = pump_fan_speed_offset
        self._temp_offset = temp_offset
        self._duty_channel = duty_channel

    def initialize(self, **kwargs):
        msg = self._request(*_REQUEST_GET_FIRMWARE)
        return [(_STATUS_FIRMWARE, "".join(map(chr, msg[3:18])), "")]

    def _get_cooler_duty(self) -> (int, int):
        """Get current pump and embedded fan duty in %."""
        msg = self._request(*_REQUEST_GET_COOLER_DUTY)
        return msg[4], msg[5]

    def _get_cooler_status(self) -> (int, int, int):
        """Get current liquid temperature, pump and embedded fan speed."""
        msg = self._request(*_REQUEST_GET_COOLER_STATUS)
        liquid_temp = msg[self._temp_offset] + msg[self._temp_offset + 1] / 10
        pump_speed = u16le_from(msg, self._pump_speed_offset)
        pump_fan_speed = u16le_from(msg, self._pump_fan_speed_offset)
        return liquid_temp, pump_speed, pump_fan_speed

    def _get_controller_speeds(self) -> List[int]:
        """Get AIO controller fan speeds in rpm."""
        msg = self._request(*_REQUEST_GET_CONTROLLER_SPEED)
        speed1 = u16le_from(msg, 5)
        speed2 = u16le_from(msg, 7)
        speed3 = u16le_from(msg, 9)
        speed4 = u16le_from(msg, 3)  # For some reason comes first in msg
        return [speed1, speed2, speed3, speed4]

    def _get_controller_duty(self) -> int:
        """Get AIO controller fan duty in %."""
        msg = self._request(*_REQUEST_GET_CONTROLLER_DUTY)
        return round(msg[4] / 0xFF * 100)

    def get_status(self, **kwargs):
        pump_duty, fan_duty = self._get_cooler_duty()
        liquid_temp, pump_speed, pump_fan_speed = self._get_cooler_status()

        status = [
            (_STATUS_TEMPERATURE, liquid_temp, "°C"),
            (_STATUS_PUMP_DUTY, pump_duty, "%"),
            (_STATUS_PUMP_SPEED, pump_speed, "rpm"),
            (_STATUS_COOLER_FAN_DUTY, fan_duty, "%"),
            (_STATUS_COOLER_FAN_SPEED, pump_fan_speed, "rpm"),
        ]

        if self._fan_count == 0:
            return status

        controller_duty = self._get_controller_duty()
        controller_speeds = self._get_controller_speeds()

        status.append((_STATUS_CONTROLLER_FAN_DUTY, controller_duty, "%"))

        for i, controller_speed in enumerate(controller_speeds):
            status.append((_STATUS_CONTROLLER_FAN_SPEED.format(i + 1), controller_speed, "rpm"))

        return status

    def _set_cooler_duties(self, pump_duty: int, fan_duty: int):
        self._write([_PREFIX, _CMD_SET_COOLER_SPEED, self._duty_channel, pump_duty, fan_duty])

    def _set_cooler_pump_duty(self, duty: int):
        pump_duty, fan_duty = self._get_cooler_duty()

        if duty == pump_duty:
            return

        self._set_cooler_duties(duty, fan_duty)

    def _set_cooler_fan_duty(self, duty: int):
        pump_duty, fan_duty = self._get_cooler_duty()

        if duty == fan_duty:
            return

        self._set_cooler_duties(pump_duty, duty)

    def _set_controller_duty(self, duty: int):
        # Controller duty is set between 0x00 and 0xFF
        duty = fraction_of_byte(percentage=duty)

        self._write([_PREFIX, _CMD_SET_CONTROLLER_SPEED, 0x00, 0x00, duty])

    def set_fixed_speed(self, channel, duty, **kwargs):
        duty = clamp(duty, 0, 100)
        channel_handlers = {
            "pump": [self._set_cooler_pump_duty],
            "pump-fan": [self._set_cooler_fan_duty],
        }

        if self._fan_count > 0:
            channel_handlers.update(
                {
                    "fans": [self._set_cooler_fan_duty, self._set_controller_duty],
                    "external-fans": [self._set_controller_duty],
                }
            )

        handlers = channel_handlers.get(channel)
        if handlers is None:
            raise ValueError(f"invalid channel: {channel}")

        for handler in handlers:
            handler(duty)

    def set_screen(self, channel, mode, value, **kwargs):
        # Not yet reverse engineered / implemented
        raise NotSupportedByDriver()

    def _request(self, request_header: int, response_header: int) -> List[int]:
        self.device.clear_enqueued_reports()
        self._write([_PREFIX, request_header])
        return self._read(response_header)

    def _read(self, expected_header=None) -> List[int]:
        msg = self.device.read(_REPORT_LENGTH)

        if msg[0] != _PREFIX:
            raise ExpectationNotMet("Unexpected report prefix")
        if expected_header is not None and msg[1] != expected_header:
            raise ExpectationNotMet("Unexpected report header")

        return msg

    def _write(self, data: List[int]):
        self.device.write(rpadlist(data, _REPORT_LENGTH, 0))



"""liquidctl drivers for Lian Li Uni hubs.

Supported devices:

- Lian Li Uni SL
- Lian Li Uni SL v2
- Lian Li Uni AL
- Lian Li Uni AL v2
- Lian Li Uni SL-Infinity

Acknowledgements:

- EightB1ts for finding IDs, PWM Commands and speed byte calculation
  https://github.com/EightB1ts/uni-sync

Copyright BlafKing and contributors
SPDX-License-Identifier: GPL-3.0-or-later
"""

import logging
import time

from enum import Enum
from liquidctl.driver.usb import UsbHidDriver
from liquidctl.util import clamp

_LOGGER = logging.getLogger(__name__)

# Minimum and maximum channel, corresponds to channels 1-4
_MIN_CHANNEL = 1
_MAX_CHANNEL = 4

# PWM commands mapped by device type
_PWM_COMMANDS = {
    "SL": [224, 16, 49],
    "AL": [224, 16, 66],
    "SLI": [224, 16, 98],
    "SLV2": [224, 16, 98],
    "ALV2": [224, 16, 98],
}

_REPORT_LENGTH = 65


class ChannelMode(Enum):
    AUTO = "auto"
    FIXED = "fixed"


class LianLiUni(UsbHidDriver):
    """Driver for Lian Li Uni fan controllers."""

    _MATCHES = [
        (0x0CF2, 0x7750, "Lian Li Uni SL", {"device_type": "SL"}),
        (0x0CF2, 0xA100, "Lian Li Uni SL", {"device_type": "SL"}),
        (0x0CF2, 0xA101, "Lian Li Uni AL", {"device_type": "AL"}),
        (0x0CF2, 0xA102, "Lian Li Uni SL-Infinity", {"device_type": "SLI"}),
        (0x0CF2, 0xA103, "Lian Li Uni SL V2", {"device_type": "SLV2"}),
        (0x0CF2, 0xA104, "Lian Li Uni AL V2", {"device_type": "ALV2"}),
        (0x0CF2, 0xA105, "Lian Li Uni SL V2", {"device_type": "SLV2"}),
    ]

    def __init__(self, device, description, **kwargs):
        """Initialize the Lian Li Uni driver."""
        super().__init__(device, description, **kwargs)
        self.device_type = kwargs.get("device_type")
        assert self.device_type, f"unexpected device with PID {hex(self.device.product_id)}"
        self.supports_cooling = True

    def initialize(self, **kwargs):
        """Initialize the device and enable PWM synchronization on all channels."""

        # Enable PWM synchronization on all channels
        for channel in range(_MIN_CHANNEL, _MAX_CHANNEL + 1):
            self.set_fan_control_mode(channel, ChannelMode.AUTO)
            time.sleep(0.2)  # Delay to prevent race conditions

        return None

    def get_status(self, **kwargs):
        """Returns queried speed from controller,
        if nothing is returned, returns 0"""
        status = []

        for channel in range(_MIN_CHANNEL, _MAX_CHANNEL + 1):
            current_speed = self._query_current_speed(channel)
            if current_speed is None:
                current_speed = 0
            status.append((f"Fan {channel} speed", int(current_speed), "rpm"))

        return status

    def set_fan_control_mode(self, channel, desired_state):
        """Toggles or explicitly sets PWM synchronization for manual speed control.

        Unstable

        Parameters:
            channel: str or (unstable) int - channel name or number ([fan]1–4);
            desired_state: ChannelMode - set AUTO to enable Auto PWM mode, FIXED to set be able to set fixed/manual speed.
        """
        channel = _parse_channel(channel)

        if desired_state is ChannelMode.AUTO:
            debug_string = "enabling"
            channel_byte = 0x11 << (channel - 1)  # enables auto mode
        else:
            debug_string = "disabling"
            channel_byte = 0x10 << (channel - 1)  # disables auto mode

        # Construct the command to toggle PWM synchronization
        command_prefix = _PWM_COMMANDS.get(self.device_type)
        assert command_prefix

        command = command_prefix + [channel_byte]
        _LOGGER.debug("%s PWM sync for channel %d: command %s", debug_string, channel, command)

        self.device.write(command)

    def set_fixed_speed(self, channel, duty, **kwargs):
        """Set a fixed speed for the specified channel.

        Parameters:
            channel: str or (unstable) int - channel name or number ([fan]1–4);
            duty: int - the desired speed percentage (0-100).
        """
        channel = _parse_channel(channel)
        duty = clamp(duty, 0, 100)

        self.set_fan_control_mode(channel, ChannelMode.FIXED)

        speed_byte = self._calculate_speed_byte(duty)
        command = [224, channel + 31, 0, speed_byte]
        _LOGGER.info("setting fan%d PWM duty to %d%%", channel, duty)
        self.device.write(command)
        time.sleep(0.1)  # Delay to prevent race conditions

    def _query_current_speed(self, channel):
        """
        Query the actual fan speed from the device for a given channel by reading the
        controller's input report.

        For device types:
          - 'SL', 'AL', 'SLI': speed data begins at offset 1.
          - 'SLV2', 'ALV2': speed data begins at offset 2.

        Each channel's speed occupies 2 bytes and is the current RPM

        Parameters:
            channel (int): channel number (1–4).

        Returns:
            int: The fan speed value as reported by the device, or raises an Assert if the read fails.
        """
        assert _MIN_CHANNEL <= channel <= _MAX_CHANNEL

        # Determine offset based on the device type.
        if self.device_type in ["SL", "AL", "SLI"]:
            offset = 1
        elif self.device_type in ["SLV2", "ALV2"]:
            offset = 2
        else:
            raise AssertionError(f"unsupported device type for reading speed: {self.device_type}")

        report = self.device.get_input_report(224, 65)

        # # Calculate the starting index for this channel's speed value.
        start_index = offset + (channel - 1) * 2
        _LOGGER.debug("start_index: %s", start_index)
        # Extract the 2 bytes corresponding to this channel.
        speed_bytes = report[start_index : start_index + 2]
        if len(speed_bytes) < 2:
            raise AssertionError(f"report is too short for channel {channel}: {report}")

        speed_value = int.from_bytes(speed_bytes, byteorder="big")
        _LOGGER.debug(
            "channel %d: extracted bytes %s -> speed value %d",
            channel,
            speed_bytes,
            speed_value,
        )

        return speed_value

    def _calculate_speed_byte(self, speed):
        """Calculate the speed byte based on the device type and desired speed.

        Parameters:
            speed: int - The desired speed percentage (0-100)

        Returns:
            int - The calculated speed byte to send to the device
        """
        if self.device_type in ["SL", "AL"]:
            if speed == 0:
                return 40
            else:
                speed_byte = int((800 + (11 * speed)) / 19) & 0xFF
        elif self.device_type == "SLI":
            if speed == 0:
                return 10
            else:
                speed_byte = int((250 + (17.5 * speed)) / 20) & 0xFF
        elif self.device_type in ["SLV2", "ALV2"]:
            if speed == 0:
                return 7
            else:
                speed_byte = int((200 + (19 * speed)) / 21) & 0xFF
        else:
            raise AssertionError(f"unsupported device type: {self.device_type}")
        return speed_byte


def _parse_channel(name_or_number):
    """Parse a channel name-or-number spec into a channel number.

    Note that passing channel numbers to the public driver APIs is an unstable feature, that may
    very well be removed.

    >>> _parse_channel("fan1")
    1

    >>> _parse_channel(1)
    1

    >>> _parse_channel("1")
    Traceback (most recent call last):
        ...
    ValueError: invalid channel '1', available channels are 'fan1'...'fan4'

    >>> _parse_channel("fan5")
    Traceback (most recent call last):
        ...
    ValueError: invalid channel 'fan5', available channels are 'fan1'...'fan4'

    >>> _parse_channel(5)
    Traceback (most recent call last):
        ...
    ValueError: invalid channel 5, available channels are 'fan1'...'fan4'
    """
    channel = None

    if isinstance(name_or_number, str) and name_or_number.startswith("fan"):
        channel = int(name_or_number.removeprefix("fan"))
    elif isinstance(name_or_number, int):
        channel = name_or_number

    if channel is None or not (_MIN_CHANNEL <= channel <= _MAX_CHANNEL):
        raise ValueError(
            f"invalid channel {name_or_number!r}, "
            f"available channels are 'fan{_MIN_CHANNEL}'...'fan{_MAX_CHANNEL}'"
        )

    return channel



"""liquidctl drivers for Lian Li Gallahad II Vision liquid coolers.

Supported devices:

- GA II LCD

Copyright Tom Frey, Jonas Malaco, Shady Nawara, Ilia Khubuluri and contributors
SPDX-License-Identifier: GPL-3.0-or-later
"""

import math
import logging
import sys

from liquidctl.driver.usb import PyUsbDevice, UsbHidDriver
from liquidctl.error import NotSupportedByDevice, NotSupportedByDriver

_LOGGER = logging.getLogger(__name__)

_PUMP_RPM_MAX = 3600
_FAN_RPM_MAX = 2520

_STATUS_TEMPERATURE = "Liquid temperature"
_STATUS_PUMP_SPEED = "Pump speed"
_STATUS_PUMP_DUTY = "Pump duty"
_STATUS_FAN_SPEED = "Fan speed"
_STATUS_FAN_DUTY = "Fan duty"

_LIGHTING_DIRECTIONS = {
    "forward": 0x00,
    "backward": 0x01,
}

_PUMP_LIGHTING_SPEEDS = {
    "slowest": 0x0,
    "slower": 0x1,
    "normal": 0x2,
    "faster": 0x3,
    "fastest": 0x4,
}

_PUMP_LIGHTING_MODES = {
    "bounce": 0x10,
    "color-morph": 0x0F,
    "burst": 0x0E,
    "big-bang": 0x0D,
    "static-starry-night": 0x0B,
    "colorful-starry-night": 0x0A,
    "transmit": 0x09,
    "fluctuation": 0x08,
    "ticker-tape": 0x07,
    "meteor": 0x06,
    "runway": 0x05,
    "breathing": 0x04,
    "static": 0x03,
    "rainbow-morph": 0x02,
    "rainbow": 0x01,
}

_FAN_LIGHTING_MODES = {
    "meteor": 0x00,
    "runway": 0x01,
    "breathing": 0x02,
    "static": 0x03,
    "rainbow-morph": 0x04,
    "rainbow": 0x05,
}

_SPEED_CHANNELS = {
    "fan": 0x8B,
    "pump": 0x8A,
}


def _quoted(*names):
    return ", ".join(map(repr, names))


def _write_colors(buffer, colors, offset, length):
    """Write variable length 3-byte RGB color arrays
    to a buffer at a specific offset."""
    buffer[offset : offset + length] = [0] * length

    if len(colors) > 0:
        buffer[offset] = colors[0][0]
        buffer[offset + 1] = colors[0][1]
        buffer[offset + 2] = colors[0][2]
    if len(colors) > 1:
        buffer[offset + 3] = colors[1][0]
        buffer[offset + 4] = colors[1][1]
        buffer[offset + 5] = colors[1][2]
    if len(colors) > 2:
        buffer[offset + 6] = colors[2][0]
        buffer[offset + 7] = colors[2][1]
        buffer[offset + 8] = colors[2][2]
    if len(colors) > 3:
        buffer[offset + 9] = colors[3][0]
        buffer[offset + 10] = colors[3][1]
        buffer[offset + 11] = colors[3][2]


class GA2LCD(UsbHidDriver):
    _status = []
    _MATCHES = [
        (
            0x0416,  # Winbond Electronics Corp.
            0x7395,  # LianLi-GA_II-LCD_v1.4
            "Lian Li GA II LCD",
            {},
        )
    ]

    def initialize(self, direct_access=False, **kwargs):
        self.device.clear_enqueued_reports()
        version = self._read_firmware_version()
        # version must contain "CA_II-Vision" somewhere in middle. Yep it's a C.
        if not ("CA_II-Vision" in version):
            raise NotSupportedByDriver(
                "Device firmware version does not match expected GA_II format: " f"{version}"
            )

        self._status.append(("Firmware version", version, ""))

        self.supports_lighting = False
        self.supports_cooling = True

        return sorted(self._status)

    def get_status(self, direct_access=False, **kwargs):
        """Get a status report.

        Returns a list of `(property, value, unit)` tuples.
        """
        self.device.clear_enqueued_reports()
        handshake = self._get_handshake()
        return [
            (_STATUS_TEMPERATURE, handshake["temperature"], "°C"),
            (_STATUS_FAN_SPEED, handshake["fan_rpm"], "rpm"),
            (_STATUS_FAN_DUTY, min(handshake["fan_rpm"] / _FAN_RPM_MAX * 100, 100), "%"),
            (_STATUS_PUMP_SPEED, handshake["pump_rpm"], "rpm"),
            (_STATUS_PUMP_DUTY, handshake["pump_rpm"] / _PUMP_RPM_MAX * 100, "%"),
        ]

    def set_color(self, channel, mode, colors, speed="normal", direction="forward", **kwargs):
        """Set the color mode for a specific channel."""

        if channel == "pump":
            self._set_pump_lighting(mode, colors, speed, direction, **kwargs)
        elif channel == "fan":
            self._set_fan_lighting(mode, colors, speed, direction, **kwargs)
        else:
            raise NotSupportedByDevice(
                f"lighting control for {channel} is not supported by this driver"
            )

    def set_speed_profile(self, channel, profile, **kwargs):
        """Not supported by this device."""
        raise NotSupportedByDevice()

    def set_fixed_speed(self, channel, duty, direct_access=False, **kwargs):
        """Set channel to a fixed speed duty."""

        if channel not in _SPEED_CHANNELS:
            raise NotSupportedByDevice(
                f"fixed speed control for {channel} is not supported by this driver"
            )

        speed = max(0, min(100, duty))
        self._write_a_cmd_with_data(_SPEED_CHANNELS[channel], [0, speed])

    def set_screen(self, channel, mode, value, **kwargs):
        """Not supported by this device."""

        # This device has a screen but it appears that it does not
        # support sending images to it, instead if relies on a
        # computer software to stream H.264 video to it over USB.
        # Liquidctl does not currently support this functionality.
        raise NotSupportedByDriver()

    def _get_a_cmd_bytes(self, cmd, data):
        # return empty list if data is empty
        if not data:
            return [
                [
                    1,  # type
                    cmd,  # command
                    0,  # reserved
                    0,  # number_a (2 bytes)
                    0,  # number_b (2 bytes)
                    0,
                ]
            ]

        num = 0
        offset = 0

        cmd_pas = []

        # take chunks of data and send them in packets until all data is sent
        while offset < len(data):
            data_len = min(len(data) - offset, 58)

            cmd_pa = [0] * 64
            cmd_pa[0] = 1
            cmd_pa[1] = cmd
            cmd_pa[2] = 0
            num_bytes = num.to_bytes(2, byteorder="big")
            cmd_pa[3] = num_bytes[0]
            cmd_pa[4] = num_bytes[1]
            cmd_pa[5] = data_len
            cmd_pa[6 : 6 + data_len] = data[offset : offset + data_len]

            _LOGGER.debug(
                "Command: 0x%02x, PDU Number: %d, Offset: %d, Data Length: %d\n",
                cmd,
                num,
                offset,
                data_len,
            )

            num += 1
            offset += data_len
            cmd_pas.append(cmd_pa)

        return cmd_pas

    def _write_a_cmd(self, cmd):
        packets = self._get_a_cmd_bytes(cmd, [])
        self._write(packets[0])

    def _write_a_cmd_with_data(self, cmd, data):
        packets = self._get_a_cmd_bytes(cmd, data)
        _LOGGER.debug("Writing command: 0x%02x", cmd)
        for packet in packets:
            self._write(packet)

    def _write(self, bytes):
        self.device.write(bytes)

    def _read_a_cmd(self):
        r = self.device.read(64)

        if not r:
            raise ValueError("Device not responding")

        if len(r) < 6:
            raise ValueError("Unexpected response from device")

        _LOGGER.debug(
            "Read command: type: 0x%02X, command: 0x%02X, number_a: 0x%02X,  number_b: 0x%02X, length: 0x%02X",
            r[0],
            r[1],
            r[3],
            r[4],
            r[5],
        )

        if r[0] != 1:
            raise ValueError("Invalid response from device")

        command = r[1]
        number = int.from_bytes(r[3:5], byteorder="big")
        length = r[5]
        current_length = min(length, 58)
        packet_count = math.ceil(length / 58)
        data = r[6 : 6 + current_length]

        if packet_count > 1:
            for i in range(packet_count - 1):
                r = self.device.read(64)
                if r[0] != 1:
                    raise ValueError("Invalid response from device")
                if r[1] != command:
                    raise ValueError("Unexpected command in response")

                number = int.from_bytes(r[3:5], byteorder="big")
                if number != i + 1:
                    raise ValueError("Unexpected response number")

                length_remaining = length - len(data)
                current_length = min(length_remaining, 58)
                data += r[6 : 6 + current_length]

        return {"command": command, "number": number, "length": length, "data": data}

    def _read_firmware_version(self):
        """Read the firmware version from the device."""
        self._write_a_cmd(0x86)

        # firmware response is two packets
        p1 = self._read_a_cmd()

        # copy until first 0
        p1_data = p1["data"][: p1["data"].index(0)] if 0 in p1["data"] else p1["data"]
        string_data = bytes(p1_data)
        ascii_string_p1 = string_data.decode("ascii", errors="ignore")

        p2 = self._read_a_cmd()
        p2_data = p2["data"][: p2["data"].index(0)] if 0 in p2["data"] else p2["data"]
        string_data = bytes(p2_data)
        ascii_string_p2 = string_data.decode("ascii", errors="ignore")
        ascii_string = ascii_string_p1.strip() + " " + ascii_string_p2.strip()

        return ascii_string

    def _get_handshake(self):
        self.device.clear_enqueued_reports()
        self._write_a_cmd(0x81)  # handshake command
        r = self._read_a_cmd()

        if r["command"] != 0x81:
            raise ValueError("Unexpected handshake response")

        if r["number"] != 0:
            raise ValueError("Unexpected handshake response number")

        if r["length"] != 0x07:
            raise ValueError("Unexpected handshake response length")

        fan_rpm = int.from_bytes(r["data"][0:2], byteorder="big")
        pump_rpm = int.from_bytes(r["data"][2:4], byteorder="big")
        temp_int = r["data"][5]
        temp_frac = r["data"][6]
        temperature = temp_int + temp_frac / 10.0

        _LOGGER.debug(
            "Handshake response: fan_rpm=%d, pump_rpm=%d, temperature=%f",
            fan_rpm,
            pump_rpm,
            temperature,
        )

        return {
            "fan_rpm": fan_rpm,
            "pump_rpm": pump_rpm,
            "temperature": temperature,
        }

    def _set_pump_lighting(self, mode, colors, speed, direction, **kwargs):
        req = [0] * 19

        if mode not in _PUMP_LIGHTING_MODES:
            raise ValueError(
                f"unknown pump lighting mode, should be one of: {_quoted(*_PUMP_LIGHTING_MODES)}"
            )

        if speed not in _PUMP_LIGHTING_SPEEDS:
            raise ValueError(
                f"unknown pump lighting speed, should be one of: {_quoted(*_PUMP_LIGHTING_SPEEDS)}"
            )

        if direction not in _LIGHTING_DIRECTIONS:
            raise ValueError(
                f"unknown lighting direction, should be one of: {_quoted(*_LIGHTING_DIRECTIONS)}"
            )

        colors = list(colors)

        req[0] = 0
        req[1] = _PUMP_LIGHTING_MODES[mode]
        req[2] = 4  # brightness 0-4
        req[3] = _PUMP_LIGHTING_SPEEDS[speed]
        _write_colors(req, colors, 4, 12)
        req[16] = _LIGHTING_DIRECTIONS[direction]
        req[17] = 0
        req[18] = 0

        self._write_a_cmd_with_data(0x83, req)

    def _set_fan_lighting(self, mode, colors, speed, direction, **kwargs):
        req = [8] * 20

        if mode not in _FAN_LIGHTING_MODES:
            raise ValueError(
                f"unknown fan lighting mode, should be one of: {_quoted(*_FAN_LIGHTING_MODES)}"
            )
        if speed not in _PUMP_LIGHTING_SPEEDS:
            raise ValueError(
                f"unknown fan lighting speed, should be one of: {_quoted(*_PUMP_LIGHTING_SPEEDS)}"
            )
        if direction not in _LIGHTING_DIRECTIONS:
            raise ValueError(
                f"unknown lighting direction, should be one of: {_quoted(*_LIGHTING_DIRECTIONS)}"
            )

        req[0] = _FAN_LIGHTING_MODES[mode]
        req[1] = 4
        req[2] = _PUMP_LIGHTING_SPEEDS[speed]
        _write_colors(req, colors, 3, 12)
        req[15] = _LIGHTING_DIRECTIONS[direction]
        req[16] = 0
        req[17] = 0
        req[18] = 0
        req[19] = 24

        self._write_a_cmd_with_data(0x85, req)
